;;; aldor.el - Aldor major mode for Emacs (version 21 or better)

;; Copyright (C) 2004 Stephen M. Wilson, all rights reserved.
;; Licensed under the GNU General Public License.

;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 2 of the License, or (at your
;; option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
;; Public License for more details.

;; You should have received a copy of the GNU General Public License along with
;; GNU Emacs; see the file COPYING.  If not, write to the Free Software
;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;; INSTALLATION
;; 
;; It is recommended that you use a modern Emacs, version 21 or better. This
;; software has been developed on FSF Emacs. There may be incompatibilities
;; when this software is used with XEmacs.
;; 
;; If you have the ability to place files in the global Emacs `site-lisp'
;; directory, then simply copy `aldor.el' into that directory. Otherwise,
;; choose a directory, move `aldor.el' there, and append a line of the
;; following form to your `.emacs':
;; 
;; 	(setq load-path (append '("foo") load-path))
;; 
;; where `foo' is the _full pathname_ to the directory where `aldor.el' can
;; be found.
;; 
;; You may also wish to add `aldor-append.el' (distributed with Aldor mode)
;; to your `.emacs' in order to have Aldor mode automatically load each
;; time you visit an Aldor source file (.as suffix).
;; 
;; If you choose not to have Aldor mode automatically load upon visiting an
;; Aldor source file, you will need to issue the command:
;; 	
;; 	M-x aldor-mode RET
;; 
;; in order to start Aldor mode.

;;; BYTE COMPILATION
;; 
;; It is highly recommended that you byte compile `aldor.el' for efficiency
;; reasons. The easiest way to do this is to open the installed `aldor.el'
;; file within Emacs, then issue the command:
;; 
;; 	M-x emacs-lisp-byte-compile RET
;; 
;; This will create a file `aldor.elc' in the installation directory -- the
;; byte compiled code for Aldor mode.


(defconst aldor-mode-version "0.8.2"
  "        Copyright (C) 2004 Stephen M. Wilson, all rights reserved.
         Copying is covered by the GNU General Public License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.")
  

;;;; Emacs version support

(defconst aldor-with-xemacs (if (string-match "XEmacs" emacs-version) t nil))
(defconst aldor-with-modern-emacs (if (string-match "21" emacs-version) t nil))


;;;; User customizable variables

;; use the standard `customize' interface or `aldor-mode-hook' to
;; configure these variables


(defgroup aldor nil
  "Support for the Aldor programming language, <http://www.aldor.org/>."
  :group 'languages )

(defcustom aldor-interactive-program "aldor -gloop"
  "*Default program name for invoking an Aldor toplevel from Emacs.")
(make-variable-buffer-local 'aldor-interactive-program)

(defcustom aldor-debugger-program "aldor -wdebugger -ldebuglib -laldor -grun"
  "*Default program name for invoking an Aldor Debugger from Emacs.")
(make-variable-buffer-local 'aldor-debugger-program)

(defcustom aldor-debugger-highlight-line t
  "*Non nil means source lines are highlighted when tracing 
with the Aldor Debugger."
  :group 'aldor :type 'boolean)

(defcustom aldor-interactive-echo-phrase t
  "*Non nil means echo input sent to toplevel Aldor buffer. 

You can toggle this setting via the command `M-x aldor-toggle-echo-phrase'."
  :group 'aldor :type 'boolean)

(defcustom aldor-pile-indentation nil
  "*Non nil turns off Aldor mode's indentation algorithm.

Mainly useful for those who use the `#pile' `#endpile' system
commands. You can toggle this setting via the command `M-x
aldor-toggle-pile-indentation'."
  :group 'aldor :type 'boolean)

(defcustom aldor-display-buffer-on-eval t
  "*Non nil means pop up the Aldor toplevel when evaluating code."
  :group 'aldor :type 'boolean)

(defcustom aldor-alert-on-error t
  "*Non nil enables audible alarms on errors."
  :group 'aldor :type 'boolean)

(defcustom aldor-toplevel-startup-code 
  '( "#include \"aldor\""
     "#include \"aldorinterp\"")
  "*Aldor code passed to the toplevel on startup.

This is a handy replacement for the `aldorinit.as' file. Each line
must be a complete Aldor expression."
  :group 'aldor :type '(repeat string))


;;; indentation defaults

(defcustom aldor-default-indent 4
  "*Default indentation.  

Global indentation variable. When no meaningful context is found,
this value is used to indent the line when required."
  :group 'aldor :type 'integer)

(defcustom aldor-indent-comments-fluidly t
  "*Non nil indents comment lines to the current code indentation
level. Otherwise, comment lines are always left justified."
  :group 'aldor :type 'boolean)

(defcustom aldor-electric-paren-indent t
  "*Non nil means electricly indent lines starting with a '}'")

(defcustom aldor-if-then-else-indent aldor-default-indent
  "*Indentation relative to an `if', `then', or `else' keyword."
  :group 'aldor :type 'integer)

(defcustom aldor-then-else-pre-indent 0
  "*Pre-indent a `this' or `else' keyword.

The indentation level of a `this' or `else' keyword is based on the
indentation level of the line containing the corresponding `if'. This
value indicates how many additional columns a `this' or `else' keyword
will be indented \(should be non-negative). For example

aldor-then-else-pre-indent = 0:

      if foo()
      then bar();
      else jar();

aldor-then-else-pre-indent = 3:

      if foo()
         then bar();
         else jar();"
  :group 'aldor :type 'integer)
      
(defcustom aldor-to-of-pre-indent aldor-default-indent
  "*Pre-indent a `to' or `of' keyword relative to `generate'.

The indentation level of a `to' or `of' keyword is based on the
indentation level of the line containing the corresponding
`generate'. This value indicates how many additional columns a `to' or
`of' keyword will be indented \(should be non-negative). For example

aldor-to-of-pre-indent = 0:

      generate 
      to foo()
      of bar();

aldor-to-of-pre-indent = 2:

      generate 
        to foo()
        of bar();"
  :group 'aldor :type 'integer)


;;; custom face definitions.

(defcustom aldor-for-pre-indent 0
  "*Pre-indent a `for' keyword relative to a companion `for'.

For example

aldor-for-pre-indent = 0;

      for i in 1..bar()
      for j in 1..jar() repeat { ... }

aldor-for-pre-indent = 4;

      for i in 1..bar()
          for j in 1..jar() repeat { ... }"
  :group 'aldor :type 'integer)


(defcustom aldor-brace-bracket-offset 1
  "*Offset between `(' and `)', or `[' and `]'.

For example

aldor-brace-bracket-offset = 2;

      foo( a,
           b
           );

aldor-brace-bracket-offset = 0;

      foo( a,
           b
         );"
  :group 'aldor :type 'integer)


(defgroup aldor-faces nil
  "Additional faces for Aldor-mode."
  :group 'aldor)

(defface aldor-font-lock-operator-face 
  '((((type tty) (class color) (background light)) (:foreground "black" :weight normal))
    (((type tty) (class color) (background dark)) (:foreground "LightGray" :weight normal))
    (((class grayscale) (background light)) (:foreground "LightGray" :bold nil))
    (((class grayscale) (background dark)) (:foreground "DimGray" :bold nil))
    (((class color) (background light)) (:foreground "CadetBlue4"))
    (((class color) (background dark)) (:foreground "CadetBlue2"))
    (t (:bold t)))
  "Aldor-mode face used to highlight operators."
  :group 'aldor-faces)


(defface aldor-font-lock-modifier-face 
  '((((type tty) (class color) (background light)) (:foreground "black" :weight bold))
    (((type tty) (class color) (background dark)) (:foreground "LightGray" :weight bold))
    (((class grayscale) (background light)) (:foreground "LightGray" :bold t))
    (((class grayscale) (background dark)) (:foreground "DimGray" :bold t))
    (((class color) (background light)) (:foreground "orchid2"))
    (((class color) (background dark)) (:foreground "orchid2"))
    (t (:bold t)))
  "Aldor-mode face used for modifier keywords."
  :group 'aldor-faces)


;;;; Utility functions for portability reasons

(when aldor-with-xemacs 
  (defun line-beginning-position (&optional count)
    (save-excursion
      (let ((num (if count count 1)))
	(beginning-of-line num)
	(point))))
  
  (defun line-end-position (&optional count)
    (save-excursion
      (let ((num (if count count 1)))
	(end-of-line count)
	(point)))))


;;;; Primitives. 

(defvar aldor-cache nil)
(make-variable-buffer-local 'aldor-cache)

(defconst aldor-comment-regexp "\\(\\+\\+\\|--\\)")

(defun aldor-find-file (f)
  (let ((buffer (find-file-noselect f 'nowarn)))
    (save-current-buffer
      (set-buffer buffer)
      (unless (eq (symbol-function major-mode)
		  (symbol-function 'aldor-mode))
	(aldor-mode)))
    buffer))
 
(defun aldor-forward-char (&optional step)
  (if step (goto-char (+ (point) step))
    (goto-char (1+ (point)))))

(defun aldor-backward-char (&optional step)
  (if step (goto-char (- (point) step))
    (goto-char (1- (point)))))

(defun aldor-in-indentation-p ()
  "Tests whether all characters between beginning of line and point
are whitespace."
  (save-excursion
    (skip-chars-backward " \t")
    (bolp)))

(defmacro even-p (n) `(= (mod ,n 2) 0))
(defmacro odd-p (n) `(= (mod ,n 2) 1))

(defun aldor-after-escape-p ()
  "Determines if point is after an escape `\_' character."
  (if (and (< (point-min) (point))
	   (char-equal ?_ (char-before)))
      (save-excursion
	(let ((savepos (point)))
	  (prog2
	      (skip-chars-backward "_")
	      (odd-p (- savepos (point))))))
    nil))

(defun aldor-in-string-p ()
  "Determines if point is in a string. If so, set the cache to the
start of the string and return true, otherwise return nil."
  (save-excursion
    (let ((lstart (aldor-start-of-line-pos))
	  (count 0)
	  (cache-it nil))
      (while (re-search-backward "\"" lstart t)
	(when (not (aldor-after-escape-p))
	  (when (not cache-it)
	    (setq cache-it (point)))
	  (setq count (1+ count))))
      (if (odd-p count)
	  (prog2
	      (setq aldor-cache cache-it) t)
	nil))))

(defun aldor-in-comment-p ()
  "Tests whether point is in a comment. If so, set the cache to the
start of the comment and return true, otherwise return nil."
  (save-excursion
    (let ((lstart (aldor-start-of-line-pos))
	  (savepos (point)))
      (while (and (re-search-backward aldor-comment-regexp lstart t)
		  (aldor-after-escape-p)) ())
      (if (and (/= savepos (point))
	       (not (aldor-after-escape-p)))
	  (if (/= lstart (point))
	      (if (aldor-in-string-p)
		  (prog2
		      (goto-char aldor-cache)
		      (aldor-in-string-or-comment-p)) ; could be just `in-comment-p'.
		(setq aldor-cache (point)) t)
	    (if (looking-at aldor-comment-regexp)
		(prog2 (setq aldor-cache (point)) t)
	      nil))
	nil))))

(defun aldor-in-string-or-comment-p ()
  (or (aldor-in-string-p)
      (aldor-in-comment-p)))

(defun aldor-end-of-line-pos ()
  "Return the position of the true end of line, accounting for escaped
newlines."
  (save-excursion
    (while (and
	    (prog2
		(end-of-line)
		(aldor-after-escape-p))
	    (< (point) (point-max)))
      (forward-line 1))
    (point)))

(defun aldor-start-of-line-pos ()	     
  "Return the position of the true start of line, accounting for escaped
newlines."
  (save-excursion
    (forward-line 0)
    (if (= (point-min) (point))
	(point)
      (while (and
	      (forward-line -1)
	      (if (= (point-min) (point))
		  (progn
		    (end-of-line)
		    (when (aldor-after-escape-p)
		      (goto-char (point-min)))
		    nil) t)
	      (prog2 (end-of-line) t)
	      (aldor-after-escape-p)))
      (unless (= (point-min) (point))
	(forward-line 1))
      (point))))

(defun aldor-alert-message (msg)
  (when aldor-alert-on-error
    (ding))
  (message msg))

(defun aldor-back-to-indentation ()
  "Like the standard back-to-indentation, but moves over lines if
spanned my a multiline string or comment."
  (back-to-indentation)
  (while (aldor-in-string-or-comment-p)
    (goto-char aldor-cache)
    (back-to-indentation)))

      
;;;; Aldor Mode Keymap

(defvar aldor-mode-map nil
  "Keymap used in Aldor mode.")
(setq aldor-mode-map (make-sparse-keymap))

(define-key aldor-mode-map "\C-c\C-r" 'aldor-eval-region)
(define-key aldor-mode-map "\C-c\C-b" 'aldor-eval-buffer)
(define-key aldor-mode-map "\t" 'aldor-indent-command)
(define-key aldor-mode-map "\C-c\C-s" 'aldor-run)
(define-key aldor-mode-map "\C-c\C-k" 'aldor-kill)
(define-key aldor-mode-map "\C-c\C-i" 'aldor-interrupt)
(define-key aldor-mode-map "\C-c\C-d" 'aldor-debugger-run)
(define-key aldor-mode-map "\C-c\C-t" 'aldor-debugger-kill)
(define-key aldor-mode-map "\C-c\C-q" 'aldor-debugger-interrupt)
(define-key aldor-mode-map "\C-cn"    'aldor-debugger-next)
(define-key aldor-mode-map "\C-cs"    'aldor-debugger-step)
(define-key aldor-mode-map "\C-cc"    'aldor-debugger-cont)
(define-key aldor-mode-map "\C-cp"    'aldor-debugger-print)
(define-key aldor-mode-map "\C-cb"    'aldor-debugger-set-breakpoint)
(define-key aldor-mode-map "\C-c\C-c" 'compile)
(define-key aldor-mode-map "}" 'aldor-electric-paren)
(define-key aldor-mode-map "\C-c\C-e" 'aldor-eval-phrase)
(define-key aldor-mode-map "\C-c\C-n" 'aldor-eval-next-phrase)
(define-key aldor-mode-map "\C-c\C-p" 'aldor-eval-previous-phrase)
(define-key aldor-mode-map "\C-c\C-j" 'aldor-beginning-of-statement)

(defvar aldor-mode-syntax-table nil
  "Syntax table in use in Aldor mode buffers.")

(defconst aldor-whitespace-syntax-descriptor 
  (if aldor-with-xemacs " " "-")
  "XEmacs favours the \" \" syntax descriptor for whitespace as
opposed to \"-\".")

(if aldor-mode-syntax-table () 
  (setq aldor-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?_ "\\" aldor-mode-syntax-table)
  (modify-syntax-entry ?- ". 124" aldor-mode-syntax-table)
  (modify-syntax-entry ?\n ">" aldor-mode-syntax-table)
  (modify-syntax-entry ?% "w" aldor-mode-syntax-table)
  (modify-syntax-entry ?? "w" aldor-mode-syntax-table)
  (modify-syntax-entry ?! "w" aldor-mode-syntax-table)
  (modify-syntax-entry ?\[ "(]" aldor-mode-syntax-table)
  (modify-syntax-entry ?\( "()" aldor-mode-syntax-table)
  (modify-syntax-entry ?\{ "(}" aldor-mode-syntax-table)
  (modify-syntax-entry ?\] ")[" aldor-mode-syntax-table)
  (modify-syntax-entry ?\) ")(" aldor-mode-syntax-table)
  (modify-syntax-entry ?\} "){" aldor-mode-syntax-table)
  (modify-syntax-entry ?\" "\"" aldor-mode-syntax-table)
  (modify-syntax-entry ?@ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?~ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?\' "." aldor-mode-syntax-table)
  (modify-syntax-entry ?\\ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?/ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?| "." aldor-mode-syntax-table)
  (modify-syntax-entry ?, "." aldor-mode-syntax-table)
  (modify-syntax-entry ?. "." aldor-mode-syntax-table)
  (modify-syntax-entry ?\; "." aldor-mode-syntax-table)
  (modify-syntax-entry ?: "." aldor-mode-syntax-table)
  (modify-syntax-entry ?^ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?# "." aldor-mode-syntax-table)
  (modify-syntax-entry ?$ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?+ "." aldor-mode-syntax-table)
  (modify-syntax-entry ?* "." aldor-mode-syntax-table)
  (modify-syntax-entry ?= "." aldor-mode-syntax-table)
  (modify-syntax-entry ?< "." aldor-mode-syntax-table)
  (modify-syntax-entry ?> "." aldor-mode-syntax-table)
  (modify-syntax-entry 
   ?\t aldor-whitespace-syntax-descriptor  aldor-mode-syntax-table)
  (modify-syntax-entry 
   ?\  aldor-whitespace-syntax-descriptor  aldor-mode-syntax-table))

  
;;;; Aldor major mode

(defun aldor-mode ()
  "Major mode for editing Aldor programs.

Provides automatic indentation and compilation. Keyword highlighting is
supported via font-lock. An Aldor Interactive mode is available as well,
allowing interaction with the Aldor Interpreter and Aldor Debugger.

Please send suggestions, questions, and bug reports to
wilsons@multiboard.com

You can access help on using Aldor mode at any time from within Emacs by
selecting the Help item from the Aldor menu or by issuing the command
`aldor-help'.

Indentation and fontification allows for Aldor code to be written easily
from within Emacs. Syntax recognition means that the standard Emacs text
movement and editing commands should work smoothly on Aldor constructs. You
can invoke the Aldor compiler, build from a Makefile, etc, by issuing the
sequence `C-c C-c'. You can subsequently visit any errors reported by Aldor
via `C-x `'

Start the Aldor Interpreter in its own Emacs buffer with the sequence `C-c
C-s', or by issuing the command `aldor-run'. You can send regions of code
from a source file directly to the Interpreter by issuing `C-c
C-r'. Alternatively, you may let Aldor Mode decide what constitutes a
logical chunk of Aldor code and issue `C-c C-e'
\(`aldor-eval-current-phrase': evaluate the current phrase around the
cursor\), `C-c C-n' \(`aldor-eval-next-phrase': evaluate the next phrase\),
or `C-c C-p' \(`aldor-eval-previous-phrase': evaluate the previous
phrase\). By design, `aldor-eval-next-phrase' and
`aldor-eval-previous-phrase' ignore system commands like `#include
\"foo.as\"'.

Start the Aldor Debugger in its own Emacs buffer with the sequence 'C-c
C-d', or by issuing the command `aldor-debugger-run'. See the debugger
documentation on how to prepare and compile files for use with the debugger
\(`aldor-debugger-run' provides a generic default command suitable for
simple situations. You may change this via the customization screen as
described in CUSTOMIZATION below\). As the debugger traces your code, the
corresponding source files are automaticly loaded into Emacs, and the
relevant line highlighted. You can remotely send `next', `step', and `cont'
commands to the debugger with the sequences `C-c n', `C-c s', and `C-c c'
respectively. Issuing `C-c p' will prompt you for an identifier name to be
used in a `print' command. Issuing `C-c b' will set a breakpoint for the
current line under the cursor. Please note that the Aldor Debugger operates
on a compiled version of your source code, thus editing your source while
the debugger is running will change the line number correspondence. This
will cause line highlighting and the setting of breakpoints to become out
of sync with the debugger. Perhaps a future version of Aldor mode will
support interactive editing/debugging.


CUSTOMIZATION

There are several parameters which can be customized. Type:

	M-x customize-group RET aldor RET

to check the defaults. Also, there are two hooks, `aldor-mode-hook' \(run
for every buffer in Aldor mode\), and `aldor-load-hook' \(run once\). These
hooks are the preferred place to add customizations of your own.

For those who write code using Aldor's `#pile' `#endpile' constructs,
automatic indentation is probably not very useful. You can turn off the
standard indentation scheme by typing:

	M-x aldor-toggle-pile-indentation RET

When Aldor-mode is asked to launch the Aldor interpreter from within Emacs,
a few lines of `startup' code are sent to the top-level by default. These
lines are:

	#include \"aldor\"
	#include \"aldorinterp\"

You can change these defaults after issuing the `customize-group' command
mentioned above, and examining the `Aldor Toplevel Startup Code' option.


SPECIAL KEYS FOR ALDOR MODE

\\{aldor-mode-map}"

  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'aldor-mode)
  (setq mode-name "Aldor")

  (use-local-map aldor-mode-map)
  (set-syntax-table aldor-mode-syntax-table)

  (make-local-variable 'paragraph-start)
  (setq paragraph-start "^[ \t]*$")
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'aldor-indent-command)
  (make-local-variable 'sentence-end)
  (setq sentence-end (concat "[;]" sentence-end))
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'comment-start)
  (setq comment-start "--")

  (aldor-build-menu)
  ;; Font-Lock support 
  (aldor-enable-font-lock)
  
  ;; hooks for aldor-mode, use them for aldor-mode configuration
  (run-hooks 'aldor-mode-hook)
  ;;(if aldor-use-abbrev-mode (abbrev-mode 1))
  (message (concat "Loading major mode for Aldor programs, version "
		   aldor-mode-version ".")))

(defun aldor-toggle-echo-phrase ()
  "Toggles on and off the echoing of phrases sent to the Aldor
toplevel."
  (interactive)
  (setq aldor-interactive-echo-phrase 
	(not aldor-interactive-echo-phrase))
  (if aldor-interactive-echo-phrase
      (message "Echo ON in Aldor toplevel.")
    (message "Echo OFF in Aldor toplevel.")))

(defun aldor-toggle-pile-indentation ()
  "Toggles between the standard Aldor-mode indentation scheme and
manual tabbing (useful for Aldor's `#pile' `#endpile' constructs.)"
  (interactive)
  (setq aldor-pile-indentation
	(not aldor-pile-indentation))
  (if aldor-pile-indentation
      (prog2
	  (define-key aldor-mode-map "\t" 'aldor-pile-indent-command)
	  (message "Pile indentation ON."))
    (define-key aldor-mode-map "\t" 'aldor-indent-command)
    (message "Pile indentation OFF.")))
    

(defun aldor-enable-font-lock ()
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(aldor-font-lock-keywords nil nil)))

		
;;;; Indentation 

(defconst aldor-keywords-regexp 
  (concat "\\<\\(a\\(dd\\|nd\\|lways\\|ssert\\)\\|b\\(reak\\|ut\\)"
	  "\\|catch\\|de\\(fault\\|fine\\)\\|do\\|e\\(lse\\|xcept\\|xtend\\)"
	  "\\|f\\(or\\|luid\\|ree\\|rom\\)\\|g\\(enerate\\|oto\\)\\|has"
	  "\\|i\\(f\\|mport\\|n\\(line\\)?\\|terate\\)\\|local\\|macro"
	  "\\|n\\(ever\\|ot\\)\\|o\\(f\\|r\\)\\|pretend"
	  "\\|re\\(f\\|peat\\|turn\\)\\|select\\|t\\(hen\\|hrow\\|o\\|ry\\)"
	  "\\|w\\(here\\|hile\\|ith\\)\\|yield\\)\\>")
  "Regexp for all recognized keywords.")


(defconst aldor-operators-regexp 
  (concat "\\*\\*\\|\\+->?\\|->\\|\\.\\.\\|/\\\\\\|:[*:]"
	  "\\|:=\\|<[<=-]\\|=\\(?:=>\\|[=>]\\)\\|>[=>]"
	  "\\|\\\\/\\|[~^]=\\|[#$'*+,./:->@\\|~^-]")
  "Regexp for all operators.")

(defconst aldor-semicolon-tie-keywords-regexp
  (concat "\\<\\(a\\(nd\\|lways\\)\\|else\\|finally\\|or\\|then\\|where"
	  "\\|\\|==\\)\\>")
  "Aldor infix keywords which can bind to statements terminated by a
`;'.")

(defconst aldor-sequence-tie-keywords-regexp
  "\\<\\(always\\|catch\\|else\\|finally\\|then\\|where\\)\\>\\|\\(==\\)"
  "Aldor infix keywords which can bind to statements terminated by a
`}'.")

(defconst aldor-group-keywords-regexp
  (concat "\\<\\(add\\|de\\(fault\\|fine\\)\\|do\\|local\\|f\\(or\\|luid\\|ree\\)"
	  "\\|generate\\|if\\|repeat\\|w\\(here\\|hile\\|ith\\)\\|export\\|try\\)"
	  "\\>\\|==")
  "Aldor keywords which can introduce a new grouping.")

(defconst aldor-implicit-group-keywords-regexp 
  (concat "\\<\\(else\\|then\\|in\\)\\>\\|" 
	  aldor-group-keywords-regexp))

(defconst aldor-syscommand-regexp 
  (concat "#\\(i\\(nclude\\(Dir\\)?\\|f\\|nt\\)"
	  "\\|e\\(lse\\(if\\)?\\|nd\\(if\\|pile\\)\\|rror\\)"
	  "\\|l\\(ibrary\\(Dir\\)?\\|ine\\)\\|\\(un\\)?assert"
	  "\\|reinclude\\|pile\\|quit\\|!"
	  "\\)\\([ \t]+\\|[ \t]*$\\)")
  "Aldor system commands.")


(defconst aldor-tie-keyword-alist
  '(("then" . "if")
    ("else" . "if")
    ("catch" . "try")
    ("finally" . "try")
    ("in" . "try\\|for"))
  "Association list of Aldor keywords which tie together.")


(defconst aldor-keyword-alist
  '(("add" . (aldor-default-indent . 0))
    ("and" . (aldor-default-indent . 0))
    ;; always deprecated, use finally.
    ("always" . (aldor-default-indent . 0))
    ("assert" . (aldor-default-indent . 0))
    ("break" . (aldor-default-indent . 0))
    ;; but deprecated, use catch.
    ("but" . (aldor-default-indent . aldor-zero-preindent))
    ("catch" . 
     (aldor-default-indent .
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "try" "catch"))))
    ("default" . (aldor-default-indent . 0))
    ("define" . (aldor-default-indent . 0))
    ("do" . (aldor-default-indent . 0))
    ("else" . 
     (aldor-if-then-else-indent .
				(lambda () 
				  (aldor-compute-corresponding-keyword-indentation
				   "if" "else" aldor-then-else-pre-indent))))
    ;; except deprecated, use throw.
    ("except" . (aldor-default-indent . 0))	
    ("export" . (aldor-default-indent . 0))
    ("extend" . (aldor-default-indent . 0))
    ("finally" . 
     (aldor-default-indent .
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "try" "finally"))))
    ("for" . (aldor-default-indent . 
				   (lambda ()
				     (aldor-compute-tight-keyword-indentation 
				      "for" "repeat" aldor-for-pre-indent))))
    ("fluid" . (aldor-default-indent . 0))
    ("free" . (aldor-default-indent . 0))
    ("from" . 
     (aldor-default-indent . 
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "import\\|inline" "from" aldor-default-indent))))
    ("generate" . (aldor-default-indent . 0))
    ("goto" . (aldor-default-indent . 0))
    ("has" . (aldor-default-indent . 0))
    ("if" . (aldor-if-then-else-indent . 0))
    ("import" . (aldor-default-indent . 0))
    ("in" . (aldor-default-indent .
				  (lambda ()
				    (aldor-compute-corresponding-keyword-indentation
				     "for\\|try" "in"))))
    ("inline" . (aldor-default-indent . 0))
    ("iterate" . (aldor-default-indent . 0))
    ("local" . (aldor-default-indent . 0))
    ("macro" . (aldor-default-indent . 0))
    ("never" . (aldor-default-indent . 0))
    ("not" . (aldor-default-indent . 0))
    ("of" . 
     (aldor-default-indent . 
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "generate" "of" aldor-to-of-pre-indent))))
    ("or" . (aldor-default-indent . 0))
    ("pretend" . (aldor-default-indent . 0))
    ("ref" . (aldor-default-indent . 0))
    ("repeat" . (aldor-default-indent . 0))
    ("return" . (aldor-default-indent . 0))
    ("select" . (aldor-default-indent . 0))
    ("then" . 
     (aldor-if-then-else-indent . 
				(lambda ()
				  (aldor-compute-corresponding-keyword-indentation
				   "if" "then" aldor-then-else-pre-indent))))
    ("throw" . (aldor-default-indent . 0))
    ("to" . 
     (aldor-default-indent .
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "generate" "to" aldor-to-of-pre-indent))))
    ("try" . (aldor-default-indent . 0))
    ("where" . (aldor-default-indent . 0))
    ("while" . (aldor-default-indent . 0))
    ("with" . (aldor-default-indent . 0))
    ("yield" . (aldor-default-indent . 0))
    ("+->" . (aldor-default-indent . 0))
    ("==" . 
     (aldor-default-indent . 
			   (lambda ()
			     (aldor-beginning-of-statement)
			     (current-column))))
    ("==>" . (aldor-default-indent . 0))
    ("=>" . (aldor-default-indent . 0))
    (":=" . (aldor-default-indent . 0))
    ("|" . 
     (aldor-default-indent . 
			   (lambda ()
			     (aldor-compute-corresponding-keyword-indentation
			      "for" "|" aldor-default-indent)))))
  "Association list of indentation values based on keywords.")


(defun aldor-syscommand-follows-p ()
  (looking-at aldor-syscommand-regexp))
      
(defun aldor-identifier-follows-p ()
  (looking-at "\\(_.\\|[a-zA-Z0-9%]\\)\\{1\\}\\(_.\\|[%!?a-zA-Z0-9]\\)*" ))

(defun aldor-comment-follows-p (&optional chunk)
  (if chunk
      (string-match "^[ \t]*\\(\\+\\+\\|--\\)" chunk)
    (looking-at "[ \t]*\\(\\+\\+\\|--\\)")))

(defun aldor-keyop-follows-p ()
  (or (looking-at aldor-keywords-regexp)
      (looking-at aldor-operators-regexp)))

(defun aldor-sequence-follows-p ()
  "Much like ALDOR-GROUP-MOD-FOLLOWS, but only checks for lines which
terminate with a `{' or a `('."
  (save-excursion
    (let ((eol (aldor-end-of-line-pos)))
      (if (and (re-search-forward ".*[({]" eol t)
	       (not (aldor-in-string-or-comment-p))
	       (setq aldor-cache (1- (point))))
	  (cond
	   ((char-equal ?\( (char-before))
	    (if (re-search-forward ")" eol t)
		(aldor-group-mod-follows-p)
	      (or (looking-at "[ \t]*$")
		  (aldor-comment-follows-p))))
	   ((char-equal ?\{ (char-before))
	    (if (re-search-forward "}" eol t)
		(aldor-group-mod-follows-p)
	      (or (looking-at "[ \t]*$")
		  (aldor-comment-follows-p)))))
	nil))))

(defun aldor-group-mod-follows-p ()
  "Checks if from point onward we have a group modifying statement;
that is, a statement followed by an opening brace.  Braces behind
comments are not recognized."
  (save-excursion 
    (let ((eol (aldor-end-of-line-pos)))
      (if (and (re-search-forward ".*[[{(]" eol t)
	       (not (aldor-in-string-or-comment-p))
	       (setq aldor-cache (1- (point))))
	  (cond
	   ((char-equal ?\( (char-before))
	    (if (re-search-forward ")" eol t)
		(aldor-group-mod-follows-p)
	      (or (looking-at "[ \t]*$")
		  (aldor-comment-follows-p))))
	   ((char-equal ?{ (char-before))
	    (if (re-search-forward "}" eol t)
		(aldor-group-mod-follows-p)
	      (or (looking-at "[ \t]*$")
		  (aldor-comment-follows-p))))
	   ((char-equal ?\[ (char-before))
	    (if (re-search-forward "]" eol t)
		(aldor-group-mod-follows-p)
	      (or (looking-at "[ \t]*$")
		  (aldor-comment-follows-p)))))
	nil))))


(defun aldor-line-ends-statement-p ()
  "Checks if the current line ends a statement. i.e., ends with a `\;'
or a `}'."
  (save-excursion
    (goto-char (aldor-end-of-line-pos))
    (if (aldor-in-string-p) nil
      (while (aldor-in-comment-p)
	(goto-char aldor-cache))
      (if (aldor-in-string-p) nil
	(skip-chars-backward " \t")
	(if (and (> (point) (point-min))
		 (or (char-equal ?\; (char-before))
		     (char-equal ?} (char-before)))
		 (setq aldor-cache (1- (point))))
	    t nil)))))

(defun aldor-sequence-ends-statement-p ()
  "Checks if the current line ends a sequence. i.e., ends with a `\)'
or a `\]'."
  (save-excursion
    (goto-char (aldor-end-of-line-pos))
    (if (aldor-in-string-p) nil
      (while (aldor-in-comment-p)
	(goto-char aldor-cache))
      (if (aldor-in-string-p) nil
	(skip-chars-backward " \t")
	(if (and (> (point) (point-min))
		 (or (char-equal ?\) (char-before))
		     (char-equal ?\] (char-before)))
		 (setq aldor-cache (1- (point))))
	    t nil)))))


(defun aldor-implicit-group-follows-p ()
  (save-excursion
    (aldor-back-to-indentation)
    (if (aldor-line-ends-statement-p) nil
      (let ((keyword nil))
	(if (save-excursion
	      (goto-char (aldor-end-of-line-pos))
	      (if (aldor-in-string-p) nil
		(while (aldor-in-comment-p)
		  (goto-char aldor-cache))
		(if (aldor-in-string-p) nil
		  (skip-chars-backward " \t")
		  (let ((lim (aldor-start-of-line-pos))
			(savepos (point)))
		    (when (and (> (point) lim)
			       (re-search-backward 
				aldor-implicit-group-keywords-regexp lim t))
		      (setq keyword (match-string-no-properties 0))
		      (and (= savepos (match-end 0))
			   (not (aldor-in-string-or-comment-p))))))))
	    keyword nil)))))
      
	
(defun aldor-skip-whitespace-comments-p (&optional compass)
  "Move point past all empty lines and all comment lines. The optional
argument specifies the direction -- if non-nil, skip backwards, else
skip forwards. Position point at the first non-blank character on the
new line, else leave point as is. Returns true if the translation was
successful, otherwise nil."
  (let ((savepos (point))
	(flag t)
	(dir (if compass -1 1)))
    (while (and flag (zerop (forward-line dir))
			    (/= (point) (point-max)))
      (unless (or (looking-at "^[ \t]*$")
		  (aldor-comment-follows-p)
		  (aldor-in-comment-p))
	(prog2 (setq flag nil) (aldor-back-to-indentation))))
    (when flag (goto-char savepos))
    (not flag)))
		     
(defun aldor-back-whitespace-comments-p ()
  "Moves back one line, skipping empty lines and comments.
Returns true if the translation was successful"
  (aldor-skip-whitespace-comments-p t))

(defun aldor-forward-whitespace-comments-p ()
  "Moves forward one line, skipping empty lines and comments.
Returns true if the translation was successful"
  (aldor-skip-whitespace-comments-p))


(defun aldor-skip-nonsystem-code-line-p (&optional compass)
  "Moves to the begining of the previous line of code, definied as
anything but a comment, blank line, or a system command such as
`#include \"aldor\"'. The optional argument specifies the direction --
if non-nil, skip backwards, else skip forwards. If a line cannot be
found, point is not changed. Returns true if the translation was
successful, otherwise nil."
  (let ((savepos (point)))
    (if (aldor-skip-whitespace-comments-p compass)
	(prog2
	    (while (and (aldor-syscommand-follows-p)
			(aldor-skip-whitespace-comments-p compass)) ())
	    (if (aldor-syscommand-follows-p)
		(prog2 (goto-char savepos) nil) t))
      nil)))

(defun aldor-back-nonsystem-code-line-p ()
  "Moves back one line, skipping comments, empty lines, and lines
starting with system command directives like `#include \"aldor\"'
Returns true if the translation was successful"
  (aldor-skip-nonsystem-code-line-p t))

(defun aldor-forward-nonsystem-code-line-p ()
  "Moves back one line, skipping comments, empty lines, and lines
starting with system command directives like `#include \"aldor\"'
Returns true if the translation was successful"
  (aldor-skip-nonsystem-code-line-p))


(defun aldor-find-matching-paren-p (paren)
  "Search for the closing parenthisis matching paren and go
there. Movement can be forwards or backwards. If no match was found
return nil, otherwise return true. If paren is an invalid argument,
signal an error."
  (if (not (string-match "[][(){}]" (char-to-string paren)))
      (error "aldor-find-matching-paren-p: Invalid argument.")
    (let ((old-pos (point)))
      (condition-case nil
	  (progn (aldor-move-out-paren-level paren) t)
	(error (progn (goto-char old-pos) nil))))))


(defun aldor-move-out-paren-level (paren)
  (goto-char (condition-case nil
		 (scan-lists (point) 
			     (if (string-match "[({[]" (char-to-string paren)) 1 -1) 
			     (if (string-match "[({[]" (char-to-string paren)) 0 1))
	       (error (point))))
  ;; Check if paren is hidden by comment or string literal.
  (when (aldor-in-string-or-comment-p)
    (goto-char aldor-cache)
    (aldor-move-out-paren-level paren)))


(defun aldor-indent-command ()
  "Indent the current line in Aldor-mode. Compute new indentation
consistent with Aldor syntax."
  (interactive "*")
  (let ((old-cfs case-fold-search))
    (setq case-fold-search nil)
    (save-excursion
      (back-to-indentation)
      (indent-line-to (aldor-compute-indentation)))
    (when (aldor-in-indentation-p) (back-to-indentation))
    (setq case-fold-search old-cfs)))


(defun aldor-pile-indent-command ()
  "Indent the current line in Aldor-mode. Always indents the line by
ALDOR-DEFAULT-INDENT spaces."
  (interactive "*")
  (save-excursion
    (aldor-back-to-indentation)
    (indent-line-to (+ (current-column) aldor-default-indent)))
  (when (aldor-in-indentation-p) (aldor-back-to-indentation)))


(defun aldor-compute-indentation ()
  (save-excursion
    (cond
     ((aldor-in-string-p) (current-column))
     ((aldor-in-comment-p) (current-column))
     ((= (point-max) (point))
      (aldor-compute-blank-line-indentation))
     ((aldor-comment-follows-p)
      (if aldor-indent-comments-fluidly
	  (aldor-compute-identifier-indentation) 0))
     ((aldor-syscommand-follows-p) 0)
     ((looking-at "[][{}()]")
      (aldor-compute-brace-indentation))
     ((and (aldor-in-brace-bracket-scope-p)
	   (or (aldor-identifier-follows-p)
	       (aldor-keyop-follows-p)
	       (looking-at "[ \t]*$")))
      (aldor-compute-brace-bracket-indentation))
     ((aldor-keyop-follows-p)
      (aldor-compute-keyop-indentation))
     ((looking-at "[ \t]*$")
      (aldor-compute-blank-line-indentation))
     (t
      (aldor-compute-identifier-indentation)))))


(defun aldor-in-brace-bracket-scope-p ()
  (save-excursion
    (let ((savepos (point)) (bracket nil))
      (when (and (aldor-find-matching-paren-p ?\))
		 (looking-at "[([]"))
	(when (/= savepos (setq bracket (point)))
	  (if (aldor-find-matching-paren-p ?\()
	      (and (< savepos (point))
		   (setq aldor-cache bracket))
	    (setq aldor-cache bracket)))))))

	    
(defun aldor-compute-brace-bracket-indentation ()
  "This should be called immediately after
aldor-in-brace-bracket-scope-p."
  (goto-char aldor-cache)
  (if (aldor-group-mod-follows-p)
      (prog2 
	  (aldor-back-to-indentation)
	  (+ (current-column) aldor-default-indent))
    (aldor-forward-char)
    (skip-chars-forward " \t")
    (current-column)))
    

(defun aldor-compute-default-indentation ()
  (if (aldor-back-whitespace-comments-p)
	(current-column)) 0)


(defun aldor-compute-indentation-after-keyword ()
  (save-excursion
    (aldor-back-to-indentation)
    (if (looking-at aldor-syscommand-regexp) 0
      (if (aldor-keyop-follows-p)
	  (let ((indent-by (assoc (match-string 0) aldor-keyword-alist)))
	    (if indent-by
		(+ (current-column) (eval (cadr indent-by)))
	      (if (aldor-line-ends-statement-p)
		  (current-column)
		(unless (aldor-implicit-group-follows-p)
		  (aldor-beginning-of-statement))
		(+ (current-column) aldor-default-indent))))
	(if (aldor-line-ends-statement-p)
	    (current-column)
	  (unless (aldor-implicit-group-follows-p)
	    (aldor-beginning-of-statement))
	  (+ (current-column) aldor-default-indent))))))


(defun aldor-compute-keyop-indentation ()
  (let ((indent-by (assoc (match-string 0) aldor-keyword-alist)))
    (if (and indent-by
	     (setq indent-by (cddr indent-by)))
	(if (functionp indent-by)
	    (funcall indent-by)
	  (progn 
	    (+ (aldor-compute-identifier-indentation) (eval indent-by))))
      (aldor-compute-identifier-indentation))))


(defun aldor-compute-tight-keyword-indentation (key &optional not-between pre-indent)
  (save-excursion
    (aldor-back-to-indentation)
    (let ((savepos (point))
	  (minpos))
      (aldor-goto-indentation-level)
      (if (= (point) savepos)
	  (aldor-compute-identifier-indentation)
	(setq minpos (point))
	(goto-char savepos)
	(when (looking-at key)
	  (aldor-backward-char))
	(if (and (re-search-backward 
		  (concat "\\<\\(" key "\\|" not-between"\\)\\>") minpos t)
		 (looking-at key))
	    (if pre-indent
		(+ (current-column) pre-indent)
	      (current-column))
	  (goto-char savepos)
	  (aldor-compute-identifier-indentation))))))

(defun aldor-compute-corresponding-keyword-indentation (parent child &optional pre-indent)
  (if (not (aldor-find-corresponding-keyword-p parent child))
      (aldor-compute-identifier-indentation)
    (aldor-back-to-indentation)
    (if pre-indent
	(+ (current-column) pre-indent)
      (current-column))))


(defun aldor-find-corresponding-keyword-p (parent child)
  (let ((savepos (point)) (count 1) (flag t)
	(match 
	 (concat "\\<\\(" parent "\\)\\>\\|}\\|"
		 (if (string-match aldor-operators-regexp child)
		     child 
		   (concat "\\<" child "\\>")))))
    (when (looking-at child)
      (while (and (/= count 0)
		  flag
		  (re-search-backward match nil t))
	(cond
	 ((aldor-in-string-or-comment-p))
	 ((looking-at child)
	  (setq count (1+ count)))
	 ((looking-at "}")
	  (unless (aldor-find-matching-paren-p ?})
	    (aldor-alert-message "Unbalanced parenthesis")
	    (setq flag nil)))
	 (t (setq count (1- count)))))
      (if (zerop count) t
	(goto-char savepos) nil))))

      
(defun aldor-compute-brace-indentation ()
  (let ((brace (char-after)))
    (cond 
     ((char-equal ?{ brace)
      (if (not (aldor-back-nonsystem-code-line-p)) 0
	(aldor-beginning-of-statement)
	(current-column)))
     ((or (char-equal ?\[ brace)
	  (char-equal ?\( brace))
      (if (aldor-in-brace-bracket-scope-p)
	  (aldor-compute-brace-bracket-indentation)
      (aldor-compute-identifier-indentation)))
     ((char-equal ?} brace)
      (if (not (aldor-find-matching-paren-p (char-after)))
	  (prog2
	      (aldor-alert-message "Unbalanced parenthesis")
	      (current-column))
	(aldor-back-to-indentation)
	(if (looking-at aldor-group-keywords-regexp)
	    (current-column)
	  (aldor-beginning-of-statement)
	  (current-column))))
     ;; We have a `\]' or a `\)'
     (t 
      (if (not (aldor-find-matching-paren-p (char-after)))
	  (prog2 
	      (aldor-alert-message "Unbalanced parenthesis")
	      (current-column))
	(+ aldor-brace-bracket-offset (current-column)))))))


(defun aldor-compute-blank-line-indentation ()
  (if (not (aldor-back-nonsystem-code-line-p)) 0
    (if (or (aldor-group-mod-follows-p)
	    (not (aldor-line-ends-statement-p))
	    (aldor-implicit-group-follows-p))
	(aldor-compute-indentation-after-keyword)
      (aldor-goto-indentation-level)
;;;   (aldor-scroll-statement)
      (current-column))))


(defun aldor-compute-identifier-indentation ()
  (let* ((savepos (point))
	 (statestart (save-excursion
		       (aldor-beginning-of-statement)
		       (point))))
    (if (not (aldor-back-nonsystem-code-line-p)) 0
      (cond 
       ((or (aldor-group-mod-follows-p)
	    (aldor-implicit-group-follows-p))
	(unless (looking-at aldor-sequence-tie-keywords-regexp)
	  (aldor-beginning-of-statement))
	(aldor-compute-indentation-after-keyword))
       ((and (= savepos statestart)
	     (aldor-line-ends-statement-p))
;;;	(aldor-scroll-statement)
	(aldor-goto-indentation-level)
	(current-column))
       ((= savepos statestart)
	(current-column))
       (t (goto-char statestart)
	  (aldor-compute-indentation-after-keyword))))))

	  
(defun aldor-scroll-statement ()
  (let ((savepos (point)) (flag nil))
    (aldor-back-to-indentation)
    (while (and (aldor-beginning-of-statement)
		(aldor-back-nonsystem-code-line-p)
		(setq flag t)
		(not (aldor-line-ends-statement-p))
		(not (aldor-group-mod-follows-p))) ())
    (when flag
      (aldor-forward-nonsystem-code-line-p)
      (when (< savepos (point))
	(goto-char savepos)))))

  
(defun aldor-current-line-number ()
  (+ (count-lines (point-min) (point))
     (if (= (current-column) 0) 1 0)))


(defun aldor-goto-indentation-level ()
  (interactive)
  (aldor-beginning-of-statement)
  (let ((prevline (aldor-current-line-number)))
    (when (aldor-back-nonsystem-code-line-p)
      (while (and (not (aldor-line-ends-statement-p))
		  (not (aldor-sequence-follows-p))
		  (if (save-excursion 
			(aldor-back-nonsystem-code-line-p))
		      (prog2 
			  (setq prevline (aldor-current-line-number))
			  (aldor-back-nonsystem-code-line-p))
		    (setq prevline (aldor-current-line-number))
		    nil))
	(aldor-beginning-of-statement))
      (goto-line prevline)))
  (aldor-back-to-indentation))
	

(defun aldor-beginning-of-statement ()
  "Find the start of the current aldor expression around point and go
there.

If point is on a line ending with a right brace `}', `)', or `]', then
the beginning of the enclosing statement is sought."
  (interactive)
  (aldor-back-to-indentation)
  (let ((savepos (point))
	(old-cfs case-fold-search))
    (setq case-fold-search nil)
    (cond
     ((= (point) (point-max))
      (aldor-beginning-of-statement-intern))
     ((when (aldor-line-ends-statement-p)
	(cond
	 ((char-equal ?} (char-after aldor-cache))
	  (goto-char aldor-cache)
	  (aldor-move-out-paren-level ?})
	  (aldor-back-to-indentation)
	  (if (= savepos (point))
	      (aldor-beginning-of-statement-intern)
	    (aldor-beginning-of-statement-intern t t))
	  t)
	 ((and (< (point-min) aldor-cache)
	       (string-match "[])]" (char-to-string (char-before aldor-cache))))
	  (goto-char (1- aldor-cache))
	  (aldor-move-out-paren-level (char-after))
	  (aldor-back-to-indentation)
	  (if (= savepos (point))
	      (aldor-beginning-of-statement-intern)
	    (aldor-beginning-of-statement-intern nil t))
	  t))))
     ((aldor-sequence-ends-statement-p)
      (goto-char aldor-cache)
      (aldor-move-out-paren-level ?})
      (aldor-back-to-indentation)
      (aldor-beginning-of-statement-intern))
     ((char-equal ?} (char-after))
      (if (aldor-find-matching-paren-p ?})
	  (prog2 
	      (aldor-back-to-indentation)
	      (aldor-beginning-of-statement-intern t t))
	(aldor-alert-message "Unbalanced parenthesis")))
     ((looking-at aldor-group-keywords-regexp) nil)
     ((looking-at aldor-sequence-tie-keywords-regexp)
      (let ((mate (assoc (match-string 0) aldor-tie-keyword-alist)))
	(when (and mate
		   (aldor-find-corresponding-keyword-p 
		    (cdr mate) (car mate)))
	  (aldor-back-to-indentation)
	  (aldor-beginning-of-statement-intern t t))))
     (t (aldor-beginning-of-statement-intern)))
    (setq case-fold-search old-cfs)
    t)) ;; always returns true


(defun aldor-beginning-of-statement-intern (&optional tie-skip delay)
  (let ((savepos (point)) (flag t) (prior-line nil) (no-skip nil))
    (while (and flag
		(if delay (or (setq delay nil) t)
		  (setq prior-line (point))
		  (aldor-back-nonsystem-code-line-p)))
      ;; We are `one line back', looking for a termination condition
      ;; which ends a statement.
      (cond
       ((aldor-line-ends-statement-p)
	;; There is a `}' or `;' at the end of this line.
	(let ((ender aldor-cache))
	  ;; Move ahead one line cautiously, making use of the cached
	  ;; value if available. Check for a tie keyword.
	  (if (save-excursion
		(when (if (and prior-line
			       (< prior-line (point-max)))
			  (goto-char prior-line)
			(aldor-forward-nonsystem-code-line-p))
		  (looking-at aldor-semicolon-tie-keywords-regexp)))
	      (when (char-equal ?} (char-after ender))
		(goto-char ender)
		(if (and (aldor-find-matching-paren-p ?})
			 (prog2 (aldor-back-to-indentation) t))
		    ;; `then', `else', etc. should keep us moving.
		    (setq tie-skip t)
		  (aldor-alert-message "Unbalanced parenthesis")
		  (setq flag nil)))
	    (setq flag nil)))) ;; No tie. 
       ((aldor-group-mod-follows-p)
	;; There is a `{', `(', or a `[' at the end of this line.
	(cond
	 ((looking-at aldor-group-keywords-regexp)
	  (setq flag nil))
	 ((and tie-skip
	       (looking-at 
		(concat "[}\)]?[ \t]*" aldor-sequence-tie-keywords-regexp)))
	  (when (looking-at "[}\)]")
	    (if (aldor-find-matching-paren-p ?})
		(aldor-back-to-indentation)
	      (aldor-alert-message "Unbalanced-parenthesis")
	      (setq flag nil))))
	 ((and  tie-skip
		(aldor-in-brace-bracket-scope-p)
		(looking-at 
		 (concat ".*\\(" aldor-group-keywords-regexp "\\)[ \t]*{"))) ())
	 (t (setq flag nil))))
       (t (let ((keyword (aldor-implicit-group-follows-p)))
	    (if keyword
		(cond
		 ((looking-at aldor-group-keywords-regexp)
		  (when tie-skip (setq no-skip t))
		  (setq flag nil))
		 ((not (and tie-skip
			    (string-match 
			     aldor-sequence-tie-keywords-regexp keyword)))
		  (setq flag nil)))
	      (when (looking-at aldor-group-keywords-regexp)
		(setq flag nil)
		(setq no-skip t)))))))
    (unless (or no-skip
		(= savepos (point))
		(and prior-line (= prior-line (point))))
      (goto-char prior-line))))


(defun aldor-electric-paren ()
  "Electrically indent parenthesis."
  (interactive "*")
  (let ((electrify (and aldor-electric-paren-indent
			(aldor-in-indentation-p)
			(not (aldor-in-string-or-comment-p)))))
    (self-insert-command 1)
    (when electrify
      (let ((old-cfs case-fold-search))
	(setq case-fold-search nil)
	(save-excursion
	  (aldor-back-to-indentation)
	  (indent-line-to (save-excursion (aldor-compute-brace-indentation))))
	(setq case-fold-search old-cfs)))))
    


;;;; Font-lock support

(defvar aldor-font-lock-keywords 
  `((,(concat "\\<\\(always\\|b\\(reak\\|ut\\)\\|catch"
	      "\\|e\\(xcept\\|lse\\)\\|f\\(inally\\|rom\\)\\|has"
	      "\\|i\\(n\\(line\\)?\\|terate\\)\\|never\\|of\\|p\\(er\\|retend\\)"
	      "\\|re\\(p\\|peat\\|f\\|turn\\)\\|t\\(hen\\|hrow\\|o\\)"
	      "\\|while\\|yield\\)\\>") 
     0 'font-lock-keyword-face nil)
    (,(concat "\\<\\(assert\\|d\\(efine\\|o\\)"
	      "\\|ex\\(port\\|tend\\)\\|for\\|generate\\|i\\(f\\|mport\\)"
	      "\\|import\\|select\\|try\\)\\>")
     0 'font-lock-function-name-face nil)
    (,(concat "\\<\\(add\\|def\\(ault\\|ine\\)\\export"
	      "\\|f\\(luid\\|ree\\)\\|local\\|w\\(here\\|ith\\)\\)\\>")
     0 'aldor-font-lock-modifier-face nil)
    ("\\<\\(goto\\|macro\\)\\>" 0 'font-lock-warning-face nil)
    ("\\<\\(and\\|not\\|or\\)\\>" 0 'font-lock-constant-face nil)
    ("[-<>=+*:$@\\/^~#|;',]\\|\\.\\."
     0 'aldor-font-lock-operator-face nil)
    (,(concat "^" aldor-syscommand-regexp) 1 'font-lock-warning-face t)
    ("\\+\\+.*$" 0 'font-lock-comment-face t))
  "Font lock patterns for Aldor mode.")


;;;; Aldor interactive mode


(require 'comint)

(defconst aldor-interactive-buffer-name "*aldor-toplevel*")
(defconst aldor-history-prompt-regexp "\\(%[1-9]+[0-9]* \\(>>\\|:=\\)\\) ")
(defconst aldor-redefine-prompt-regexp "\\(Redefine\\? (y/n):\\) ")
(defconst aldor-continuation-prompt-regexp "\\.\\.\\. ")
(defconst aldor-history-promptln-regexp 
  (concat aldor-history-prompt-regexp "[ \t]*$"))
(defconst aldor-redefine-promptln-regexp 
  (concat aldor-redefine-prompt-regexp "[ \t]*$"))
(defconst aldor-continuation-promptln-regexp 
  (concat aldor-continuation-prompt-regexp "[ \t]*$"))
(defconst aldor-interactive-wait-mseconds 0)
(defconst aldor-interactive-wait-seconds 1)
(defvar aldor-in-sensitive-indentation-mode nil)
(defvar aldor-interactive-mode-map nil)
(if aldor-interactive-mode-map nil
  (setq aldor-interactive-mode-map
        (copy-keymap comint-mode-map)))
(define-key aldor-interactive-mode-map "\C-c\C-k" 'aldor-kill)
(define-key aldor-interactive-mode-map "\C-c\C-i" 'aldor-interrupt)


(defadvice yank (around aldor-interactive-yank-advice)
  "Yanks can corrupt the Aldor toplevel as they can deposit several
lines at once into the buffer, causing the interpreters line-by-line
responses to `pile up' at the end of the input. This piece of advice
completely overrides the yank command when called in the Aldor
toplevel."
  (if (not (or (string-equal 
		(buffer-name (current-buffer)) 
		aldor-interactive-buffer-name)
	       (string-equal
		(buffer-name (current-buffer))
		aldor-debugger-buffer-name)))
      ad-do-it
    (let* ((arg (ad-get-arg 0))
	   (ring-elt (if (numberp arg) arg 0))
	   (chunk (current-kill ring-elt))
	   (len (length chunk))
	   (chunk-start 0) (code-line nil))
      (while (and (< chunk-start len)
		  (string-match ".*$" chunk chunk-start))
	(setq code-line (match-string 0 chunk))
	(setq chunk-start (+ chunk-start (length code-line) 1))
	(when (not (string-match "^[ \t]*$" code-line))
	  (set-text-properties 0 (length code-line) nil code-line)
	  (insert code-line)
	  (when (or (\= len chunk-start)
		    (char-equal ?\n (char-before)))
		    (comint-send-input)
		    (when (not aldor-in-sensitive-indentation-mode)
		      (aldor-expect-toplevel-output))))))
    (setq this-command 'yank)
    (setq ad-return-value nil)))


(defun aldor-interactive-mode ()
  "Major mode for interacting with the Aldor Interpreter. Runs the
Aldor Interpreter as an Emacs subprocess, with Input/Output though an
Emacs buffer. There is a command line history, and code can be sent
from another buffer in Aldor mode.

Special keys for Aldor Interactive mode:

\\{aldor-interactive-mode-map}"
  (interactive)
  (message (concat "Loading Aldor interactive mode, version "
		   aldor-mode-version "."))
  (comint-mode)
  (setq comint-process-echoes nil)
  (setq major-mode 'aldor-interactive-mode)
  (setq mode-name "Aldor-Interactive")
  (setq comint-scroll-to-bottom-on-output t)
  (setq comint-scroll-to-bottom-on-input t)
  (make-local-variable 'comint-prompt-regexp)
  (setq comint-prompt-regexp
	(concat aldor-history-prompt-regexp "\\|" aldor-redefine-prompt-regexp))
  (use-local-map aldor-interactive-mode-map)
  (make-local-hook 'comint-input-filter-functions)
  (add-hook 'comint-input-filter-functions 'aldor-check-indent-mode)
  (set-syntax-table aldor-mode-syntax-table)
  (ad-activate 'yank)
  ;; We dont have abbrevs yet. 
  ;;  (setq local-abbrev-table aldor-mode-abbrev-table)
  (run-hooks 'aldor-interactive-mode-hook))


(defun aldor-run (&optional cmd)
  "Run an Aldor toplevel process. I/O via buffer named `*aldor-toplevel*'."
  (interactive
   (list (if (not (comint-check-proc aldor-interactive-buffer-name))
	     (read-from-minibuffer "Aldor toplevel to run: "
				   aldor-interactive-program))))
  (aldor-run-process-if-needed cmd)
  (when aldor-display-buffer-on-eval
    (display-buffer aldor-interactive-buffer-name)))


(defun aldor-run-process-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc aldor-interactive-buffer-name)
	  (setq cmd aldor-interactive-program)
	(setq cmd (read-from-minibuffer "Aldor toplevel to run: "
					aldor-interactive-program))))
  (setq aldor-interactive-program cmd)
  (if (not (comint-check-proc aldor-interactive-buffer-name))
      (let ((cmdlist (aldor-args-to-list cmd)))
	(set-buffer (apply (function make-comint) "aldor-toplevel"
			   (car cmdlist) nil (cdr cmdlist)))
	(aldor-interactive-mode)
	(sleep-for 2)
	(aldor-interactive-startup-code))))


(defun aldor-args-to-list (string)
  (let ((where (string-match "[ \t]" string)))
    (cond ((null where) (list string))
	  ((not (= where 0))
	   (cons (substring string 0 where)
		 (aldor-args-to-list (substring string (+ 1 where)
						(length string)))))
	  (t (let ((pos (string-match "[^ \t]" string)))
	       (if (null pos)
		   nil
		 (aldor-args-to-list (substring string pos
						(length string)))))))))


(defun aldor-interactive-startup-code ()
  "Passes code to the Aldor toplevel at startup. Intended
mostly for things like `#include \"aldor\"' and `#include \"aldorinterp\"'"
  (when (boundp 'aldor-toplevel-startup-code)
    (mapcar #'(lambda (chunk)
		(if aldor-interactive-echo-phrase
		    (save-excursion
		      (set-buffer aldor-interactive-buffer-name)
		      (goto-char (point-max))
		      (insert chunk)
		      (goto-char (point-max))
		      (comint-send-input))
		  (aldor-simple-send chunk))
		(aldor-expect-toplevel-output))
	    aldor-toplevel-startup-code)))


(defun aldor-find-statement()
  (save-excursion
    (let ((start (save-excursion
		   (aldor-goto-indentation-level)
;;;		   (aldor-scroll-statement)
		   (point)))
	  (no-skip nil) (flag t))
      (when (and (= start (point))
		 (or (looking-at "[ \t]*$")
		     (aldor-comment-follows-p)))
	(aldor-forward-whitespace-comments-p)
	(setq start (point)))
      (aldor-back-to-indentation)
      (unless (and (aldor-syscommand-follows-p)
		   (= start (point))
		   (setq no-skip t))
    	(while flag 
	  (cond 
	   ((aldor-line-ends-statement-p)
	    (if (aldor-forward-nonsystem-code-line-p)
		(unless (looking-at aldor-sequence-tie-keywords-regexp)
		  (setq flag nil))
	      (setq no-skip t)
	      (setq flag nil)))
	   ((aldor-group-mod-follows-p)
	    (goto-char aldor-cache)
	    (aldor-find-matching-paren-p (char-after)))
	   (t
	    (if (aldor-forward-nonsystem-code-line-p)
		(when (char-equal ?} (char-after))
		  (setq flag nil))
	      (setq no-skip t)
	      (setq flag nil))))))
      (unless no-skip
	(aldor-back-nonsystem-code-line-p))
      (goto-char (aldor-end-of-line-pos))
      (cons start (point)))))

(defun aldor-eval-region (start end)
  "Evaluate the current region in the Aldor toplevel."
  (interactive "r")
  (save-excursion (aldor-run-process-if-needed))
  (comint-preinput-scroll-to-bottom)
  (when aldor-display-buffer-on-eval
    (display-buffer aldor-interactive-buffer-name))
  (when (< end start)
    (let ((tmp start))
      (setq start end)
      (setq end tmp)))
  (save-excursion
    (goto-char start)
    (while (< start end)
      (let* ((chunk-end (line-end-position))
	     (text (buffer-substring-no-properties start chunk-end)))
	(when (and (not (string-match "^[ \t]*$" text))
		   (not (string= text ""))
		   (not (aldor-comment-follows-p text)))
	  (save-excursion
	    (set-buffer aldor-interactive-buffer-name)
	    (goto-char (point-max))
	    (when (and (not aldor-interactive-echo-phrase)
		       (eq 'history (aldor-get-toplevel-prompt-type)))
	      ;; Output at least one newline to keep the prompts consistent.
	      (insert "\n")
	      (comint-send-input))
	    (if aldor-interactive-echo-phrase
		(progn
		  (insert text)
		  (comint-send-input))
	      (aldor-simple-send text)))
	  (when (not aldor-in-sensitive-indentation-mode)
	    (aldor-expect-toplevel-output))))
      (setq start (line-beginning-position 2))
      (goto-char start)))
  (when aldor-display-buffer-on-eval
    (display-buffer aldor-interactive-buffer-name)))

(defun aldor-eval-phrase ()
  "Evaluate the current phrase in the Aldor toplevel."
  (interactive)
  (let ((tuple (aldor-find-statement)))
    (aldor-eval-region (car tuple) (cdr tuple))))


(defun aldor-eval-next-phrase ()
  "Evaluate the next phrase (if any) in the Aldor toplevel. The search
for the next phrase skips past Aldor system command statements such as
`#include \"aldor\"'"
  (interactive)
  (let ((savepos (point))
	(tuple (aldor-find-statement)))
    (aldor-back-to-indentation)
    (unless (< (point) (car tuple))
      (goto-char (cdr tuple)))
    (if (aldor-forward-nonsystem-code-line-p)
	(let ((tuple (aldor-find-statement)))
	  (aldor-eval-region (car tuple) (cdr tuple)))
      (goto-char savepos)
      (aldor-alert-message "Next phrase not found"))))

(defun aldor-eval-previous-phrase ()
  "Evaluate the previous phrase (if any) in the Aldor toplevel. The
search for the next phrase skips past Aldor system command statements
such as `#include \"aldor\"'"
  (interactive)
  (let ((savepos (point)))
    (goto-char (car (aldor-find-statement)))
    (if (aldor-back-nonsystem-code-line-p)
	(let ((tuple (aldor-find-statement)))
	  (aldor-eval-region (car tuple) (cdr tuple)))
      (goto-char savepos)
      (aldor-alert-message "No previous phrase found."))))


(defun aldor-simple-send (chunk)
  "Like COMINT-SIMPLE-SEND, but calls the input hooks as well."
  (run-hook-with-args 'comint-input-filter-functions
		      (concat chunk "\n"))
  (comint-simple-send aldor-interactive-buffer-name chunk))


(defun aldor-expect-toplevel-output (&optional dont-wait)
  (save-excursion
    (set-buffer aldor-interactive-buffer-name)
    (let ((flag t) (count 0))
      (while (and flag (< count 3))
	(1+ count)
	(if (or dont-wait
		(aldor-wait-toplevel-output-p))
	    ;; We want MAX-POS assigned as soon as possible, 
	    (let* ((max-pos (point-max))
		   (prompt (aldor-get-toplevel-prompt-type)))
	      (goto-char (point-max))
	      (forward-line 0)
	      (if (eval 'prompt)
		  (cond
		   ((eq 'continuation prompt)
		    (unless aldor-interactive-echo-phrase
		      (delete-region (point) (line-end-position)))
		    (setq flag nil))
		   ((eq 'redefine prompt)
		    ;; Changed for compatibility with XEmacs
		    ;;(comint-bol)
		    (comint-bol nil)
		    (let ((ans (y-or-n-p "Redefine? ")))
		      (if ans 
			  (insert "Y")
			(insert "N"))
		      (comint-send-input)
		      (setq dont-wait nil)))
		   ((eq 'history prompt) 
		    ;; Changed for compatibility with XEmacs
		    ;;(comint-bol)
		    (comint-bol nil)
		    (setq flag nil)))
		(when (< max-pos (point-max))
		  (setq dont-wait t)))))
	(setq dont-wait nil)))))


(defun aldor-get-toplevel-prompt-type ()
  "If point is at an Aldor prompt, return its type, else nil.
Types are: `history', `redefine', `continuation'."
  (save-excursion
    (set-buffer aldor-interactive-buffer-name)
    (goto-char (point-max))
    (forward-line 0)
    (cond
     ((looking-at aldor-continuation-promptln-regexp) 'continuation)
     ((looking-at aldor-history-promptln-regexp) 'history)
     ((looking-at aldor-redefine-promptln-regexp) 'redefine)
     (t nil))))


(defun aldor-check-indent-mode (chunk)
  "Updates the current indentation mode given the current line 
of code seen by the interpreter."
  (when (not (string= chunk ""))
    (if aldor-in-sensitive-indentation-mode
	(if (cdr aldor-in-sensitive-indentation-mode)
	    (when (string-match "^#endif[ \t].*$" chunk)
	      (setq aldor-in-sensitive-indentation-mode nil))
	  (when (or (not (eq ?\t (string-to-char chunk)))
		    (not (eq ?\ (string-to-char chunk))))
	    (setq aldor-in-sensitive-indentation-mode nil)))
      ;; Newlines on their own send aldor into indent-sensitive mode,
      ;; regardless of context.
      (cond
       ((string-match "^#if[ \t].*$" chunk)
	(setq aldor-in-sensitive-indentation-mode (cons t t)))
       ((and (string-match "^.*==[ \t]*$" chunk)
	     (not (string-match
		   (concat "^.*" aldor-comment-regexp ".*==[ \t]*$") chunk)))
	(setq aldor-in-sensitive-indentation-mode (list t)))))))


(defun aldor-wait-toplevel-output-p ()
  (let ((proc (get-buffer-process aldor-interactive-buffer-name)))
    (when (eq 'run (process-status proc))
      (accept-process-output proc
			     aldor-interactive-wait-seconds
			     aldor-interactive-wait-mseconds))))


(defun aldor-eval-buffer ()
  "Send the buffer to the Aldor toplevel."
  (interactive)
  (aldor-eval-region (point-min) (point-max)))


(defun aldor-kill ()
  "Kill the Aldor toplevel process."
  (interactive)
  (when (comint-check-proc aldor-interactive-buffer-name)
    (save-excursion
      (set-buffer aldor-interactive-buffer-name)
      (comint-kill-subjob))))


(defun aldor-interrupt ()
  "Quit the Aldor toplevel process."
  (interactive)
  (when (comint-check-proc aldor-interactive-buffer-name)
    (save-excursion
      (set-buffer aldor-interactive-buffer-name)
      (comint-send-eof)
      (when (comint-check-proc aldor-interactive-buffer-name)
	(comint-kill-subjob)))))



;;;; Aldor Debugger Mode



(defconst aldor-debugger-buffer-name "*aldor-debugger*")
(defconst aldor-debugger-fileinfo-regexp 
  "\\[\"\\(.*\\)\" at line \\([0-9]+\\)\\]")
(defconst aldor-debug-prompt-regexp "\\((debug)\\) ")
(defvar aldor-debugger-overlay nil)
(defvar aldor-debugger-mode-map nil)
(if aldor-debugger-mode-map nil
  (setq aldor-debugger-mode-map
        (copy-keymap comint-mode-map)))
(define-key aldor-debugger-mode-map "\C-c\C-k" 'aldor-debugger-kill)
(define-key aldor-debugger-mode-map "\C-c\C-i" 'aldor-debugger-interrupt)


(defun aldor-debugger-mode ()
  "Major mode for interacting with the Aldor Debugger. Runs the
Aldor Debugger as an Emacs subprocess, with Input/Output though an
Emacs buffer. There is a command line history, automated stepping 
through Aldor source files, and remote control of the Debugger from
another buffer in Aldor mode.

Special keys for Aldor Debugger mode:

\\{aldor-debugger-mode-map}"
  (interactive)
  (message (concat "Loading Aldor Debugger mode, version "
		   aldor-mode-version "."))
  (comint-mode)
  (setq comint-process-echoes nil)
  (setq major-mode 'aldor-debugger-mode)
  (setq mode-name "Aldor-Debugger")
  (make-local-variable 'comint-prompt-regexp)
  (setq comint-prompt-regexp aldor-debug-prompt-regexp)
  (make-local-hook 'comint-output-filter-functions)
  (add-hook 'comint-output-filter-functions 
	    'aldor-debugger-output-filter)
  (setq comint-scroll-to-bottom-on-output t)
  (setq comint-scroll-to-bottom-on-input t)
  (use-local-map aldor-interactive-mode-map)
  (set-syntax-table aldor-mode-syntax-table)
  (ad-activate 'yank)
  (run-hooks 'aldor-debugger-mode-hook))
  

(defun aldor-debugger-run (&optional cmd)
  "Run an Aldor toplevel process. I/O via buffer named `*aldor-debugger*'."
  (interactive
   (list (unless (comint-check-proc aldor-debugger-buffer-name)
	   (read-from-minibuffer "Aldor debugger to run: "
				 (concat aldor-debugger-program " "
					 (file-relative-name buffer-file-name))))))
  (aldor-debugger-run-if-needed cmd)
  (when aldor-display-buffer-on-eval
    (display-buffer aldor-debugger-buffer-name)))


(defun aldor-debugger-output-filter (chunk)
  (when (string-match aldor-debugger-fileinfo-regexp chunk)
    (let ((fname (match-string-no-properties 1 chunk))
	  (lineno (match-string-no-properties 2 chunk)))
      (when (and (file-readable-p fname)
		 (stringp fname)
		 (stringp lineno))
	(aldor-debugger-display-line 
	 fname (string-to-number lineno))))))


(defun aldor-debugger-run-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc aldor-debugger-buffer-name)
	  (setq cmd aldor-debugger-program)
	(setq cmd (read-from-minibuffer "Aldor debugger to run: " 
				 (concat aldor-debugger-program " "
					 (file-relative-name buffer-file-name))))))
  (if (not (comint-check-proc aldor-debugger-buffer-name))
      (let ((cmdlist (aldor-args-to-list cmd)))
	(set-buffer (apply (function make-comint) "aldor-debugger"
			   (car cmdlist) nil (cdr cmdlist)))
	(aldor-debugger-mode)
	(sit-for 2)
	(when (comint-check-proc aldor-debugger-buffer-name)
	  (set-process-sentinel (get-buffer-process aldor-debugger-buffer-name)
				'aldor-debugger-sentinel)))))


(defun aldor-debugger-unhighlight-line ()
  (when aldor-debugger-overlay
    (delete-overlay aldor-debugger-overlay)))


(defun aldor-debugger-highlight-line ()
  (aldor-debugger-unhighlight-line)
  (save-excursion
    (forward-line 0)
    (setq aldor-debugger-overlay 
	  (make-overlay (point) (aldor-end-of-line-pos)))
    (overlay-put aldor-debugger-overlay 'face 'highlight)))


(defun aldor-debugger-clean-highlighting ()
  (when aldor-debugger-overlay
    (save-current-buffer 
      (set-buffer (overlay-buffer aldor-debugger-overlay))
      (aldor-debugger-unhighlight-line)
      (set-marker overlay-arrow-position nil))))


(defun aldor-debugger-display-line (the-file line)
  (let* ((buffer
          (save-current-buffer
            (or (eq (current-buffer) aldor-debugger-buffer-name)
                (set-buffer aldor-debugger-buffer-name))
            (aldor-find-file the-file)))
         (window 
	  (and buffer 
	       (or (get-buffer-window buffer)
		   (prog2
		       (if (eq (selected-window) 
			       (get-buffer-window aldor-debugger-buffer-name))
			   (display-buffer buffer)
		       (set-window-buffer (selected-window) buffer)
		       (display-buffer buffer))))))
         (pos))
    (if buffer
        (progn
	  (save-current-buffer 
	    (save-excursion
	      (set-buffer buffer)
	      (save-restriction
		(widen)
		(goto-line line)
		(when aldor-debugger-highlight-line
		  (aldor-debugger-highlight-line))
		(setq pos (point))
		(setq overlay-arrow-string "=>")
		(or overlay-arrow-position
		    (setq overlay-arrow-position (make-marker)))
		(set-marker overlay-arrow-position (point) (current-buffer)))
	      (cond ((or (< pos (point-min)) (> pos (point-max)))
		     (widen)
		     (goto-char pos))))
	    (set-window-point window overlay-arrow-position))))))


(defun aldor-debugger-send (chunk)
  (save-excursion
    (set-buffer aldor-debugger-buffer-name)
    (comint-goto-process-mark)
    (insert chunk)
    (delete-region (point) (point-max))
    (comint-send-input)))


(defun aldor-debugger-next ()
  "Send a `next' command to the Aldor Debugger."
  (interactive)
  (if (comint-check-proc aldor-debugger-buffer-name)
      (prog2
	  (unless (eq (current-buffer) aldor-debugger-buffer-name)
	    (display-buffer aldor-debugger-buffer-name t))
	  (aldor-debugger-send "next"))
    (aldor-alert-message "No debugger process active.")))


(defun aldor-debugger-step ()
  "Send a `step' command to the Aldor Debugger."
  (interactive)
  (if (comint-check-proc aldor-debugger-buffer-name)
      (prog2
	  (unless (eq (current-buffer) aldor-debugger-buffer-name)
	    (display-buffer aldor-debugger-buffer-name t))
	  (aldor-debugger-send "step"))
    (aldor-alert-message "No debugger process active.")))


(defun aldor-debugger-cont ()
  "Send a `cont' command to the Aldor Debugger."
  (interactive)
  (if (comint-check-proc aldor-debugger-buffer-name)
      (prog2 
	  (unless (eq (current-buffer) aldor-debugger-buffer-name)
	    (display-buffer aldor-debugger-buffer-name t))
	  (aldor-debugger-send "cont"))
    (aldor-alert-message "No debugger process active.")))


(defun aldor-debugger-print ()
  "Prompts for a variable name whos value the Aldor Debugger
is then asked  to print."
  (interactive)
  (if (comint-check-proc aldor-debugger-buffer-name)
      (progn
	(let ((matp minibuffer-allow-text-properties)
	      (var))
	  (setq minibuffer-allow-text-properties nil)
	  (setq var (read-from-minibuffer "print: "))
	  (unless (eq (current-buffer) aldor-debugger-buffer-name)
	    (display-buffer aldor-debugger-buffer-name t))
	  (aldor-debugger-send (concat "print " var))
	  (setq minibuffer-allow-text-properties matp)))
    (aldor-alert-message "No debugger process active.")))

  
(defun aldor-debugger-set-breakpoint ()
  "Sets the current line as a breakpoint in the Aldor Debugger."
  (interactive)
  (if (comint-check-proc aldor-debugger-buffer-name)
      (let ((dbgdir (save-current-buffer
		      (set-buffer aldor-debugger-buffer-name)
		      (expand-file-name default-directory))))
	(unless (eq (current-buffer) aldor-debugger-buffer-name)
	  (display-buffer aldor-debugger-buffer-name t))
	(aldor-debugger-send 
	 (concat "break "
		 (if (string= (file-name-directory buffer-file-name) dbgdir)
		     (file-name-nondirectory buffer-file-name)
		   buffer-file-name) " "
		   (int-to-string (+ (count-lines (point-min) (point))
				     (if (= (current-column) 0) 1 0))))))
    (aldor-alert-message "No debugger process active.")))
  

(defun aldor-debugger-sentinel (process event)
  (save-current-buffer 
    (set-buffer aldor-debugger-buffer-name)
    (aldor-debugger-clean-highlighting)
    (goto-char (point-max))
    (insert
     (format "\nProcess %s %s" (process-name process) event))))


(defun aldor-debugger-kill ()
  "Kill the Aldor debugger process."
  (interactive)
  (when (comint-check-proc aldor-debugger-buffer-name)
    (save-excursion
      (set-buffer aldor-debugger-buffer-name)
      (comint-kill-subjob))))


(defun aldor-debugger-interrupt ()
  "Quit the Aldor debugger process."
  (interactive)
  (when (comint-check-proc aldor-debugger-buffer-name)
    (save-excursion
      (set-buffer aldor-debugger-buffer-name)
      (comint-send-eof)
      (when (comint-check-proc aldor-debugger-buffer-name)
	(comint-kill-subjob)))))



;;;; Menu support

(defun aldor-about ()
  (interactive)
  (describe-variable 'aldor-mode-version))

(defun aldor-help ()
  (interactive)
  (describe-function 'aldor-mode))

(defun aldor-interactive-help ()
  "Get help on the Aldor Toplevel."
  (interactive)
  (describe-function 'aldor-interactive-mode))

(defun aldor-debugger-help ()
  "Get help on the Aldor Debugger."
  (interactive)
  (describe-function 'aldor-debugger-mode))


(defun aldor-build-menu ()
  (if (condition-case nil (prog2 (require 'easymenu) nil) (error t)) ()
    (easy-menu-define 
      aldor-mode-menu (list aldor-mode-map)
      "Aldor Mode Menu"
      '("Aldor"
	("Interactive Mode"
	 ["Run Aldor Toplevel" aldor-run t]
	 ["Interupt Aldor Toplevel" aldor-interrupt
	  :active (comint-check-proc aldor-interactive-buffer-name)]
	 ["Kill Aldor Toplevel" aldor-kill
	  :active (comint-check-proc aldor-interactive-buffer-name)]
	 ["Help" aldor-interactive-help t])
	("Debugger Mode"
	 ["Run Aldor Debugger" aldor-debugger-run t]
	 ["Set break point" aldor-debugger-set-breakpoint
	  :active (comint-check-proc aldor-debugger-buffer-name)]
	 ["Interupt Aldor Debugger" aldor-debugger-interrupt
	  :active (comint-check-proc aldor-debugger-buffer-name)]
	 ["Kill Aldor Debugger" aldor-debugger-kill
	  :active (comint-check-proc aldor-debugger-buffer-name)]
	 ["Help" aldor-debugger-help t])
	["Compile..." compile t]
	["Evaluate Region" aldor-eval-region
	 :active (if (fboundp 'region-active-p) (region-active-p) mark-active)]
	["Comment Out Region" comment-region
	 :active (if (fboundp 'region-active-p) (region-active-p) mark-active)]
	["Uncomment Region" uncomment-region
	 :active (if (fboundp 'region-active-p) (region-active-p) mark-active)]
	"---"
	["Customize Aldor Mode" (customize-group 'aldor) t]
	["About" aldor-about t]
	["Help" aldor-help t]))
    (easy-menu-add aldor-mode-menu)))
    

;;;; Hooks and Exit

(defvar aldor-load-hook nil
  "This hook is run when Aldor mode is loaded in. This is the best
place to put key-bindings and your hacks upon the Font-Lock keywords")

(defvar aldor-mode-hook nil
  "This hook is run every time a buffer enters Aldor mode. This is a
good place run turn-on-font-lock.")

(defvar aldor-interactive-mode-hook nil
  "This hook is run every time a buffer enters Aldor Interactive mode.")

(defvar aldor-debugger-mode-hook nil
  "This hook is run every time a buffer enters Aldor Debugger mode.")

(run-hooks 'aldor-load-hook)

(provide 'aldor)
